@@ -36,7 +36,7 @@
 
 static void pg_get_one_r(char *value, Oid arg_out_fn_oid, SEXP *obj,
 																int elnum);
-static SEXP get_r_vector(Oid typtype, int numels);
+static SEXP get_r_vector(Oid typtype, int64 numels);
 static Datum get_trigger_tuple(SEXP rval, plr_function *function,
 									FunctionCallInfo fcinfo, bool *isnull);
 static Datum get_tuplestore(SEXP rval, plr_function *function,
@@ -52,19 +52,17 @@ static Datum get_generic_array_datum(SEXP rval, plr_function *function, int col,
 static Tuplestorestate *get_frame_tuplestore(SEXP rval,
 											 plr_function *function,
 											 AttInMetadata *attinmeta,
-											 MemoryContext per_query_ctx,
-											 bool retset);
+											 MemoryContext per_query_ctx);
 static Tuplestorestate *get_matrix_tuplestore(SEXP rval,
 											 plr_function *function,
 											 AttInMetadata *attinmeta,
-											 MemoryContext per_query_ctx,
-											 bool retset);
+											 MemoryContext per_query_ctx);
 static Tuplestorestate *get_generic_tuplestore(SEXP rval,
 											 plr_function *function,
 											 AttInMetadata *attinmeta,
-											 MemoryContext per_query_ctx,
-											 bool retset);
+											 MemoryContext per_query_ctx);
 static SEXP coerce_to_char(SEXP rval);
+static Datum r_get_tuple(SEXP rval, plr_function *function, FunctionCallInfo fcinfo);
 
 extern char *last_R_error_msg;
 
@@ -324,130 +322,93 @@ pg_array_get_r(Datum dvalue, FmgrInfo out_func, int typlen, bool typbyval, char
 	return result;
 }
 
+#ifdef HAVE_WINDOW_FUNCTIONS
 /*
- * Given an array pg datums, convert to a multi-row R vector.
+ * Evaluate a window function's argument expression on a specified
+ *		window frame, returning R array for the argno column in the frame
+ *
+ * winobj: PostgreSQL window object handle
+ * argno: argument number to evaluate (counted from 0)
+ * function: contains necessary info on how to output Datum as string for general case conversion
  */
 SEXP
-pg_datum_array_get_r(Datum *elem_values, bool *elem_nulls, int numels, bool has_nulls,
-					 Oid element_type, FmgrInfo out_func, bool typbyval)
+pg_window_frame_get_r(WindowObject winobj, int argno, plr_function* function)
 {
 	/*
 	 * Loop through and convert each scalar value.
 	 * Use the converted values to build an R vector.
 	 */
 	SEXP		result;
-	int			i;
-	bool		fast_track_type;
-
-	switch (element_type)
-	{
-		case INT4OID:
-		case FLOAT8OID:
-			fast_track_type = true;
-			break;
-		default:
-			fast_track_type = false;
-	}
+	int			numels = 0;
+	Oid			element_type = function->arg_typid[argno];
+	FmgrInfo	out_func = function->arg_out_func[argno];
+	int64		totalrows = WinGetPartitionRowCount(winobj);
 
 	/*
-	 * Special case for pass-by-value data types, if the following conditions are met:
-	 * 		designated fast_track_type
-	 * 		no NULL elements
-	 * 		1 dimensional array only
-	 * 		at least one element
+	 * Get new vector of the appropriate type.
+	 * We presume unbound frame as a common use case in R.
+	 * If not, we will trim vector later.
 	 */
-	if (fast_track_type &&
-		 typbyval &&
-		 !has_nulls &&
-		 (numels > 0))
+	PROTECT(result = get_r_vector(element_type, totalrows));
+
+	/* Convert all values to their R form and build the vector */
+	for (;; numels++)
 	{
-		SEXP	matrix_dims;
+		char	   *value;
+		bool		isnull;
+		Datum		dvalue;
+		bool		isout = false;
+		bool		set_mark = (0 == numels);
 
-		/* get new vector of the appropriate type and length */
-		PROTECT(result = get_r_vector(element_type, numels));
+		dvalue = WinGetFuncArgInFrame(winobj, argno, numels, WINDOW_SEEK_HEAD,
+			set_mark, &isnull, &isout);
+
+		if (isout)
+			break;
 
-		/* keep this in sync with switch above -- fast_track_type only */
 		switch (element_type)
 		{
+			case BOOLOID:
+				LOGICAL_DATA(result)[numels] = isnull ? NA_LOGICAL : DatumGetBool(dvalue);
+				break;
+			case INT8OID:
+				NUMERIC_DATA(result)[numels] = isnull ? NA_REAL : (double)DatumGetInt64(dvalue);
+				break;
+			case INT2OID:
 			case INT4OID:
-				Assert(sizeof(int) == 4);
-				memcpy(INTEGER_DATA(result), elem_values, numels * sizeof(int));
+			case OIDOID:
+				INTEGER_DATA(result)[numels] = isnull ? NA_INTEGER : DatumGetInt32(dvalue);
+				break;
+			case FLOAT4OID:
+				NUMERIC_DATA(result)[numels] = isnull ? NA_REAL : DatumGetFloat4(dvalue);
 				break;
 			case FLOAT8OID:
-				Assert(sizeof(double) == 8);
-				memcpy(NUMERIC_DATA(result), elem_values, numels * sizeof(double));
+				NUMERIC_DATA(result)[numels] = isnull ? NA_REAL : DatumGetFloat8(dvalue);
 				break;
 			default:
-				/* Everything else is error */
-				ereport(ERROR,
-						(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-						 errmsg("direct array passthrough attempted for unsupported type")));
+				value = isnull ? NULL : DatumGetCString(FunctionCall3(&out_func,
+															dvalue,
+															(Datum) 0,
+															Int32GetDatum(-1)));
+
+				/*
+					* Note that pg_get_one_r() replaces NULL values with
+					* the NA value appropriate for the data type.
+					*/
+				pg_get_one_r(value, element_type, &result, numels);
+				if (value != NULL)
+					pfree(value);
 		}
-
-		/* attach dimensions */
-		PROTECT(matrix_dims = allocVector(INTSXP, 1));
-		INTEGER_DATA(matrix_dims)[0] = numels;
-		setAttrib(result, R_DimSymbol, matrix_dims);
-		UNPROTECT(1);
-
-		UNPROTECT(1);	/* result */
 	}
-	else
-	{
-		SEXP	matrix_dims;
 
-		/* array is empty */
-		if (numels == 0)
-		{
-			PROTECT(result = get_r_vector(element_type, 0));
-			UNPROTECT(1);
+	if (numels != totalrows)
+		SET_LENGTH(result, numels);
 
-			return result;
-		}
-
-		/* get new vector of the appropriate type and length */
-		PROTECT(result = get_r_vector(element_type, numels));
-
-		/* Convert all values to their R form and build the vector */
-		for (i = 0; i < numels; i++)
-		{
-			char	   *value;
-			Datum		itemvalue;
-			bool		isnull;
-
-			isnull = elem_nulls[i];
-			itemvalue = elem_values[i];
-
-			if (!isnull)
-			{
-				value = DatumGetCString(FunctionCall3(&out_func,
-													  itemvalue,
-													  (Datum) 0,
-													  Int32GetDatum(-1)));
-			}
-			else
-				value = NULL;
-
-			/*
-			 * Note that pg_get_one_r() replaces NULL values with
-			 * the NA value appropriate for the data type.
-			 */
-			pg_get_one_r(value, element_type, &result, i);
-			if (value != NULL)
-				pfree(value);
-		}
-
-		/* attach dimensions */
-		PROTECT(matrix_dims = allocVector(INTSXP, 1));
-		INTEGER_DATA(matrix_dims)[0] = numels;
-		setAttrib(result, R_DimSymbol, matrix_dims);
-		UNPROTECT(1);
-
-		UNPROTECT(1);	/* result */
-	}
+	UNPROTECT(1);	/* result */
 
 	return result;
 }
+#endif
 
 /*
  * Given an array of pg tuples, convert to an R list
@@ -586,7 +547,7 @@ pg_tuple_get_r_frame(int ntuples, HeapTuple *tuples, TupleDesc tupdesc)
  * create an R vector of a given type and size based on pg output function oid
  */
 static SEXP
-get_r_vector(Oid typtype, int numels)
+get_r_vector(Oid typtype, int64 numels)
 {
 	SEXP	result;
 
@@ -653,7 +614,12 @@ pg_get_one_r(char *value, Oid typtype, SEXP *obj, int elnum)
 			 * because R INTEGER is only 4 byte
 			 */
 			if (value)
-				NUMERIC_DATA(*obj)[elnum] = atof(value);
+			{
+				/* fixup for Visual Studio 2013, _MSC_VER == 1916*/
+				char *endptr = NULL;
+				const double el = strtod(value, &endptr);
+				NUMERIC_DATA(*obj)[elnum] = value==endptr ? R_NaN : el;
+			}
 			else
 				NUMERIC_DATA(*obj)[elnum] = NA_REAL;
 			break;
@@ -681,21 +647,12 @@ r_get_pg(SEXP rval, plr_function *function, FunctionCallInfo fcinfo)
 	bool	isnull = false;
 	Datum	result;
 
-	if (function->result_typid != BYTEAOID &&
-		(TYPEOF(rval) == CLOSXP ||
-		 TYPEOF(rval) == PROMSXP ||
-		 TYPEOF(rval) == LANGSXP ||
-		 TYPEOF(rval) == ENVSXP))
-		ereport(ERROR,
-				(errcode(ERRCODE_DATA_EXCEPTION),
-				 errmsg("incorrect function return type"),
-				 errdetail("R return value type cannot be mapped to PostgreSQL return type."),
-				 errhint("Try BYTEA as the PostgreSQL return type.")));
-
 	if (CALLED_AS_TRIGGER(fcinfo))
 		result = get_trigger_tuple(rval, function, fcinfo, &isnull);
-	else if (function->result_istuple || fcinfo->flinfo->fn_retset)
+	else if (fcinfo->flinfo->fn_retset)
 		result = get_tuplestore(rval, function, fcinfo, &isnull);
+	else if (function->result_natts > 1)
+		result = r_get_tuple(rval, function, fcinfo);
 	else
 	{
 		/* short circuit if return value is Null */
@@ -705,8 +662,8 @@ r_get_pg(SEXP rval, plr_function *function, FunctionCallInfo fcinfo)
 			return (Datum) 0;
 		}
 
-		if (function->result_elem == 0)
-			result = get_scalar_datum(rval, function->result_typid, function->result_in_func, &isnull);
+		if (function->result_fld_elem_typid[0] == function->result_fld_typid[0])
+			result = get_scalar_datum(rval, function->result_fld_typid[0], function->result_fld_elem_in_func[0], &isnull);
 		else
 			result = get_array_datum(rval, function, 0, &isnull);
 
@@ -718,6 +675,53 @@ r_get_pg(SEXP rval, plr_function *function, FunctionCallInfo fcinfo)
 	return result;
 }
 
+/*
+ * Given an R value (data frame or list), coerce it to list
+ * and get a tuple representing first elements of each list element.
+ *
+ * This is used to return a single RECORD (not SETOF)
+ */
+Datum
+r_get_tuple(SEXP rval, plr_function *function, FunctionCallInfo fcinfo)
+{
+	Oid			oid;
+	TupleDesc	tupdesc;
+	HeapTuple	tuple;
+	Datum	   *values;
+	bool	   *isnull;
+	int			i, min_length;
+
+	if (!(isFrame(rval) || isNewList(rval) || isList(rval)))
+		elog(ERROR, "Only list alike is expected");
+
+	if (TYPEFUNC_COMPOSITE != get_call_result_type(fcinfo, &oid, &tupdesc))
+		elog(ERROR, "return type must be a row type");
+
+	min_length = Min(function->result_natts, length(rval));
+
+	//if (tupdesc->natts != length(rval))
+	//	elog(ERROR, "same length expected");
+
+	BlessTupleDesc(tupdesc);
+
+	values = palloc0(sizeof(Datum) * tupdesc->natts);
+	isnull = palloc0(sizeof(bool) * tupdesc->natts);
+
+	for (i = 0; i < min_length; i++)
+	{
+		SEXP el = VECTOR_ELT(rval, i);
+		if (function->result_fld_typid[i] != function->result_fld_elem_typid[i])
+			values[i] = get_array_datum(el, function, i, isnull + i);
+		else
+			values[i] = get_scalar_datum(el, function->result_fld_elem_typid[i], function->result_fld_elem_in_func[i], isnull + i);
+	}
+
+	tuple = heap_form_tuple(tupdesc, values, isnull);
+	pfree(values);
+	pfree(isnull);
+	return HeapTupleGetDatum(tuple);
+}
+
 /*
  * Similar to r_get_pg, given an R value, convert to its pg representation
  * Other than scalar, currently only prepared to be used with simple 1D vector
@@ -926,7 +930,6 @@ get_trigger_tuple(SEXP rval, plr_function *function, FunctionCallInfo fcinfo, bo
 static Datum
 get_tuplestore(SEXP rval, plr_function *function, FunctionCallInfo fcinfo, bool *isnull)
 {
-	bool			retset = fcinfo->flinfo->fn_retset;
 	ReturnSetInfo  *rsinfo = (ReturnSetInfo *) fcinfo->resultinfo;
 	TupleDesc		tupdesc;
 	AttInMetadata  *attinmeta;
@@ -975,14 +978,12 @@ get_tuplestore(SEXP rval, plr_function *function, FunctionCallInfo fcinfo, bool
 	/* OK, go to work */
 	rsinfo->returnMode = SFRM_Materialize;
 
-	if (isFrame(rval))
-		rsinfo->setResult = get_frame_tuplestore(rval, function, attinmeta, per_query_ctx, retset);
-	else if (isList(rval) || isNewList(rval))
-		rsinfo->setResult = get_frame_tuplestore(rval, function, attinmeta, per_query_ctx, retset);
+	if (isFrame(rval) || isList(rval) || isNewList(rval))
+		rsinfo->setResult = get_frame_tuplestore(rval, function, attinmeta, per_query_ctx);
 	else if (isMatrix(rval))
-		rsinfo->setResult = get_matrix_tuplestore(rval, function, attinmeta, per_query_ctx, retset);
+		rsinfo->setResult = get_matrix_tuplestore(rval, function, attinmeta, per_query_ctx);
 	else
-		rsinfo->setResult = get_generic_tuplestore(rval, function, attinmeta, per_query_ctx, retset);
+		rsinfo->setResult = get_generic_tuplestore(rval, function, attinmeta, per_query_ctx);
 
 	/*
 	 * SFRM_Materialize mode expects us to return a NULL Datum. The actual
@@ -1130,7 +1131,7 @@ get_array_datum(SEXP rval, plr_function *function, int col, bool *isnull)
 	else
 	{
 		/* create an empty array */
-		return PointerGetDatum(construct_empty_array(function->result_elem));
+		return PointerGetDatum(construct_empty_array(function->result_fld_elem_typid[col]));
 	}
 }
 
@@ -1165,22 +1166,11 @@ get_frame_array_datum(SEXP rval, plr_function *function, int col, bool *isnull)
 		/* internal error */
 		elog(ERROR, "plr: bad internal representation of data.frame");
 
-	if (function->result_istuple)
-	{
-		result_elem = function->result_fld_elem_typid[col];
-		in_func = function->result_fld_elem_in_func[col];
-		typlen = function->result_fld_elem_typlen[col];
-		typbyval = function->result_fld_elem_typbyval[col];
-		typalign = function->result_fld_elem_typalign[col];
-	}
-	else
-	{
-		result_elem = function->result_elem;
-		in_func = function->result_elem_in_func;
-		typlen = function->result_elem_typlen;
-		typbyval = function->result_elem_typbyval;
-		typalign = function->result_elem_typalign;
-	}
+	result_elem = function->result_fld_elem_typid[col];
+	in_func = function->result_fld_elem_in_func[col];
+	typlen = function->result_fld_elem_typlen[col];
+	typbyval = function->result_fld_elem_typbyval[col];
+	typalign = function->result_fld_elem_typalign[col];
 	
 	for (j = 0; j < nc; j++)
 	{
@@ -1367,7 +1357,7 @@ get_md_array_datum(SEXP rval, int ndims, plr_function *function, int col, bool *
 	int			cntr = 0;
 	bool	   *nulls;
 	bool		have_nulls = FALSE;
-	Oid 		return_type_oid = function->result_elem;
+	Oid 		return_type_oid = function->result_fld_elem_typid[col];
 
 	if (ndims > 0)
 	{
@@ -1380,22 +1370,11 @@ get_md_array_datum(SEXP rval, int ndims, plr_function *function, int col, bool *
 		lbs = NULL;
 	}
 	
-	if (function->result_istuple)
-	{
-		result_elem = function->result_fld_elem_typid[col];
-		in_func = function->result_fld_elem_in_func[col];
-		typlen = function->result_fld_elem_typlen[col];
-		typbyval = function->result_fld_elem_typbyval[col];
-		typalign = function->result_fld_elem_typalign[col];
-	}
-	else
-	{
-		result_elem = function->result_elem;
-		in_func = function->result_elem_in_func;
-		typlen = function->result_elem_typlen;
-		typbyval = function->result_elem_typbyval;
-		typalign = function->result_elem_typalign;
-	}
+	result_elem = function->result_fld_elem_typid[col];
+	in_func = function->result_fld_elem_in_func[col];
+	typlen = function->result_fld_elem_typlen[col];
+	typbyval = function->result_fld_elem_typbyval[col];
+	typalign = function->result_fld_elem_typalign[col];
 
 	PROTECT(rdims = getAttrib(rval, R_DimSymbol));
 	for(i = 0; i < ndims; i++)
@@ -1522,7 +1501,7 @@ get_md_array_datum(SEXP rval, int ndims, plr_function *function, int col, bool *
 					else
 					{
 						nulls[arridx] = FALSE;
-						dvalues[arridx] = Float8GetDatum((double) REAL(rval)[idx]);
+						dvalues[arridx] = DirectFunctionCall1(float8_numeric, Float8GetDatum((double)REAL(rval)[idx]));
 					}
 				}
 			}
@@ -1646,22 +1625,11 @@ get_generic_array_datum(SEXP rval, plr_function *function, int col, bool *isnull
 	bool		fast_track_type;
 	bool		has_na = false;
 
-	if (function->result_istuple)
-	{
-		result_elem = function->result_fld_elem_typid[col];
-		in_func = function->result_fld_elem_in_func[col];
-		typlen = function->result_fld_elem_typlen[col];
-		typbyval = function->result_fld_elem_typbyval[col];
-		typalign = function->result_fld_elem_typalign[col];
-	}
-	else
-	{
-		result_elem = function->result_elem;
-		in_func = function->result_elem_in_func;
-		typlen = function->result_elem_typlen;
-		typbyval = function->result_elem_typbyval;
-		typalign = function->result_elem_typalign;
-	}
+	result_elem = function->result_fld_elem_typid[col];
+	in_func = function->result_fld_elem_in_func[col];
+	typlen = function->result_fld_elem_typlen[col];
+	typbyval = function->result_fld_elem_typbyval[col];
+	typalign = function->result_fld_elem_typalign[col];
 
 	/*
 	 * Special case for pass-by-value data types, if the following conditions are met:
@@ -1805,7 +1773,7 @@ get_generic_array_datum(SEXP rval, plr_function *function, int col, bool *isnull
 				else
 				{
 					nulls[i] = FALSE;
-					dvalues[i] = Float8GetDatum((double) REAL(rval)[i]);
+					dvalues[i] = DirectFunctionCall1(float8_numeric, Float8GetDatum((double)REAL(rval)[i]));
 				}
 			}
 		}
@@ -1870,8 +1838,7 @@ static Tuplestorestate *
 get_frame_tuplestore(SEXP rval,
 					 plr_function *function,
 					 AttInMetadata *attinmeta,
-					 MemoryContext per_query_ctx,
-					 bool retset)
+					 MemoryContext per_query_ctx)
 {
 	Tuplestorestate	   *tupstore;
 	char			  **values;
@@ -1904,18 +1871,13 @@ get_frame_tuplestore(SEXP rval,
 	 * If we return a set, get number of rows by examining the first column.
 	 * Otherwise, stop at one row.
 	 */
-	if (retset)
+	if (isFrame(rval))
 	{
-		if (isFrame(rval))
-		{
-			PROTECT(dfcol = VECTOR_ELT(rval, 0));
-			nr = length(dfcol);
-			UNPROTECT(1);
-		}
-		else if (isList(rval) || isNewList(rval))
-			nr = 1;
+		PROTECT(dfcol = VECTOR_ELT(rval, 0));
+		nr = length(dfcol);
+		UNPROTECT(1);
 	}
-	else
+	else if (isList(rval) || isNewList(rval))
 		nr = 1;
 
 	/* coerce columns to character in advance */
@@ -1933,25 +1895,6 @@ get_frame_tuplestore(SEXP rval,
 			SET_VECTOR_ELT(result, j, obj);
 			UNPROTECT(1);
 		}
-		else if(TUPLE_DESC_ATTR(tupdesc,j)->attndims != 0)	/* array data type */
-		{
-			SEXP	obj;
-
-			PROTECT(obj = NEW_LIST(nr));
-			for(i = 0; i < nr; i++)
-			{
-				SEXP	dfcolcell;
-				SEXP	objcell;
-
-				PROTECT(dfcolcell = VECTOR_ELT(dfcol, i));
-				PROTECT(objcell = coerce_to_char(dfcolcell));
-				SET_VECTOR_ELT(obj, i, objcell);
-				UNPROTECT(2);
-			}
-
-			SET_VECTOR_ELT(result, j, obj);
-			UNPROTECT(1);
-		}
 		else
 		{
 			SEXP 	t;
@@ -2007,55 +1950,10 @@ get_frame_tuplestore(SEXP rval,
 				else
 					values[j] = NULL;
 			}
+			else if (STRING_ELT(dfcol, i) != NA_STRING)
+				values[j] = pstrdup(CHAR(STRING_ELT(dfcol, i)));
 			else
-			{
-				if ((TUPLE_DESC_ATTR(tupdesc,j)->attndims != 0) || (STRING_ELT(dfcol, i) != NA_STRING))
-				{
-					if (TUPLE_DESC_ATTR(tupdesc,j)->attndims == 0)
-					{
-						values[j] = pstrdup(CHAR(STRING_ELT(dfcol, i)));
-					}
-					else	/* array data type */
-					{
-						bool	isnull = false;
-						Datum	arr_datum;
-
-						if (TYPEOF(dfcol) != VECSXP)
-							arr_datum = get_array_datum(dfcol, function, j, &isnull);
-						else
-							arr_datum = get_array_datum(VECTOR_ELT(dfcol,i), function, j, &isnull);
-
-						if (isnull)
-						{
-							values[j] = NULL;
-						}
-						else
-						{
-							FunctionCallInfoData	fake_fcinfo;
-							FmgrInfo				flinfo;
-							Datum					dvalue;
-							
-							MemSet(&fake_fcinfo, 0, sizeof(fake_fcinfo));
-							MemSet(&flinfo, 0, sizeof(flinfo));
-							fake_fcinfo.flinfo = &flinfo;
-							flinfo.fn_mcxt = CurrentMemoryContext;
-							fake_fcinfo.context = NULL;
-							fake_fcinfo.resultinfo = NULL;
-							fake_fcinfo.isnull = false;
-							fake_fcinfo.nargs = 1;
-							fake_fcinfo.arg[0] = arr_datum;
-							fake_fcinfo.argnull[0] = false;
-							dvalue = (*array_out)(&fake_fcinfo);
-							if (fake_fcinfo.isnull)
-								values[j] = NULL;
-							else
-								values[j] = DatumGetCString(dvalue);
-						}
-					}
-				}
-				else
-					values[j] = NULL;
-			}
+				values[j] = NULL;
 
 			UNPROTECT(1);
 		}
@@ -2089,8 +1987,7 @@ static Tuplestorestate *
 get_matrix_tuplestore(SEXP rval,
 					 plr_function *function,
 					 AttInMetadata *attinmeta,
-					 MemoryContext per_query_ctx,
-					 bool retset)
+					 MemoryContext per_query_ctx)
 {
 	Tuplestorestate	   *tupstore;
 	char			  **values;
@@ -2108,10 +2005,7 @@ get_matrix_tuplestore(SEXP rval,
 	 * If we return a set, get number of rows.
 	 * Otherwise, stop at one row.
 	 */
-	if (retset)
-		nr = nrows(rval);
-	else
-		nr = 1;
+	nr = nrows(rval);
 
 	/* initialize our tuplestore */
 	tupstore = TUPLESTORE_BEGIN_HEAP;
@@ -2156,8 +2050,7 @@ static Tuplestorestate *
 get_generic_tuplestore(SEXP rval,
 					 plr_function *function,
 					 AttInMetadata *attinmeta,
-					 MemoryContext per_query_ctx,
-					 bool retset)
+					 MemoryContext per_query_ctx)
 {
 	Tuplestorestate	   *tupstore;
 	char			  **values;
@@ -2175,10 +2068,7 @@ get_generic_tuplestore(SEXP rval,
 	 * If we return a set, get number of rows.
 	 * Otherwise, stop at one row.
 	 */
-	if (retset)
-		nr = length(rval);
-	else
-		nr = 1;
+	nr = length(rval);
 
 	/* initialize our tuplestore */
 	tupstore = TUPLESTORE_BEGIN_HEAP;
