@@ -52,15 +52,15 @@ typedef struct saved_plan_desc
  * Functions used in R
  *****************************************************************************/
 void
-throw_pg_notice(const char **msg)
+throw_pg_log(int* elevel, const char **msg)
 {
 	/* skip error CONTEXT for explicitly called messages */
 	SAVE_PLERRCONTEXT;
 
 	if (msg && *msg)
-		elog(NOTICE, "%s", *msg);
+		elog(*elevel, "%s", *msg);
 	else
-		elog(NOTICE, "%s", "");
+		elog(*elevel, "%s", "");
 
 	RESTORE_PLERRCONTEXT;
 }
@@ -597,6 +597,7 @@ plr_SPI_execp(SEXP rsaved_plan, SEXP rargvalues)
 	return result;
 }
 
+#if CATALOG_VERSION_NO < 201811201
 /*
  * plr_SPI_lastoid - return the last oid. To be used after insert queries.
  */
@@ -611,6 +612,7 @@ plr_SPI_lastoid(void)
 
 	return result;
 }
+#endif
 
 /*
  * Takes the prepared plan rsaved_plan and creates a cursor 
